# pylama: max_line_length = 100
from enum import IntEnum, unique
# import traceback

from PyQt5.QtWidgets import \
    QWidget, QLabel, QPushButton, QRadioButton, QButtonGroup
from PyQt5.QtCore import QTimer

from mydatetime import DateWidget, ClockWidget
from info import InfoWidget
from alert import AlertWidget
from menu import MenuWidget
from others.reason import ReasonWidget
from others.balance import BalanceWidget
from brightness import Brightness
from clock import Clock

from others.screentest import ScreenTestWidget

@unique
class EventCode(IntEnum):
    """Codes, returned by events from s30.

    Пожалуйста, убедитесь лишний раз, что они совпадают с кодами в regevent.h
    """

    IDLOCK_NOTINLIST = 1
    IDLOCK_NOVALID = 2
    PASS_PERSONAL = 11
    IDENTIFY_FAIL = 12
    IDENTIFY_FP_VERIFY_FAIL = 13
    IDENTIFY_FP_VERIFY = 0x59
    INDICATION_EVENT_OFF = 0x5A


class Direction(IntEnum):
    """Codes of passage direction, send to and received from s30."""

    NONE = -1
    ENTER = 0
    EXIT = 1


class MainWindow(QWidget):

    @unique
    class State(IntEnum):
        NONE = 0
        ENTER = 1
        EXIT = 2
        PRE_MENU = 3
        MENU = 4
        PRE_BALANCE = 5
        BALANCE = 6
        WAIT_FOR_REASON = 7
        REASON = 8
        ASK_TO_CHOOSE = 9

    CONVERT_STATE_TO_DIR = {
        State.NONE: Direction.NONE,
        State.ENTER: Direction.ENTER,
        State.EXIT: Direction.EXIT,
    }

    _state = None
    state_default = State.NONE

    def __init__(self, protocol, parent=None):
        super().__init__(parent)
        self.resize(800, 480)
        pic = QLabel(self)
        pic.setGeometry(0, 0, 800, 480)
        pic.setObjectName('main')

        protocol.events.connect(self._on_event_answer)
        protocol.users.connect(self.userAnswer)
        protocol.answer.connect(self.protocolAnswer)
        protocol.connected.connect(self._on_socket_connect)
        protocol.disconnected.connect(self._on_socket_disconnect)
        self.protocol = protocol

        self.brightness = Brightness()
        self.clock = Clock()

        self.reason_text = None

        self.makeCenter()
        self.makeButtons()
        # self.buttonsClear()

        menu_open = QPushButton(None, self)
        menu_open.setObjectName('menu_open')
#        menu_open.clicked.connect(self.menuOpenClick)
        menu_open.clicked.connect(lambda: self.set_state(self.State.PRE_MENU))
        self.menu_open = menu_open

        date = DateWidget(self.clock, self)
        date.setGeometry(662, 30, 771-662, 45-30)

        self.setLang()

        # timer = QTimer(self)
        # timer.timeout.connect(self.getDefault)
        # timer.start(1000 * 60)
        # self.timer = timer
        #
        # self.timer_reason = None

        # self.popup = None
        # self.msg = None
        self.menu = MenuWidget(self.protocol, self)
        self.menu.go_back.connect(self.go_to_default_state)
        self.balance = BalanceWidget(self)
        self.balance.go_back.connect(self.go_to_default_state)
        self.reason = ReasonWidget(self.protocol, self)
        self.reason.accept.connect(self.setReason)
        self.reason.cancel.connect(self.go_to_default_state)

        tstwidg = ScreenTestWidget(protocol, self)

        self.info_msg = InfoWidget('', self)
        self.alert_msg = AlertWidget('', self)
        self.clockMsg('')
        timer = QTimer(self)
        timer.timeout.connect(self.go_to_default_state)
        timer.setInterval(1000 * 10)
        timer.setSingleShot(True)
        self.card_waiting_timeout = timer

        # self.go_to_default_state()
        self.reason_state = self.State.NONE
        self.go_to_default_state()

        self.clock.min.connect(self.protocol.getDefault)

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, value):
        self.set_state(value)

    def set_state(self, value, keep_info_on_screen=False):
        if self._state is value:
            return
        if value is self.State.WAIT_FOR_REASON:
            if self.group.checkedId() in (self.State.ENTER, self.State.EXIT):
                self.reason_state = self.group.checkedId()
                self.reason_text = None
            else:
                value = self.State.ASK_TO_CHOOSE

        print('State changed from:', self._state, ' to ', value,
              ' (', keep_info_on_screen, ')')
        self._state = value
        if value == self.State.MENU:
            self.menu.show()
            return
        if value == self.State.BALANCE:
            self.card_waiting_timeout.stop()
            return
        if value == self.State.WAIT_FOR_REASON:
            self.card_waiting_timeout.stop()
            self.reason.show()
            return

        if value in (self.State.NONE, self.State.ENTER, self.State.EXIT):
            self.group.button(value).setChecked(True)
            self.protocol.defaultDir(self.CONVERT_STATE_TO_DIR[value])

        if value == self.State.NONE:
            self.clockMsg('', keep_info_on_screen)
        elif value == self.State.ASK_TO_CHOOSE:
            self.clockMsg(_("Выберите \"%s\" или \"%s\"") %
                          (_("Вход"), _("Выход")))
        elif value in (self.State.ENTER, self.State.EXIT):
            self.clockMsg(
                _("Поднесите карту или приложите палец к сканеру"),
                keep_info_on_screen)
        elif value == self.State.PRE_MENU:
            self.infoMsg(
                _("Для входа в меню<br>предъявите палец или карту"))
        elif value == self.State.PRE_BALANCE:
            self.infoMsg(
                _("Для получения баланса<br>предъявите палец или карту"))
        elif value == self.State.REASON:
            self.infoMsg(_("Поднесите карту или приложите палец к сканеру<br>"
                           "для регистрации причины отсутствия"))
        self.card_waiting_timeout.start()

    def go_to_default_state(self, keep_info_on_screen=False):
        self.set_state(self.state_default, keep_info_on_screen)
        self.protocol.getDefault()

    def makeCenter(self):
        clock = ClockWidget(self)
        clock.setGeometry(257, 148, 541-257, 204-148)
        self.clock_wdg = clock

        label = QLabel(None, self)
        label.setGeometry(0, 230, 800, 32)
        label.setObjectName('center_msg')
        self.clock_msg = label
        self.popup = None

    def makeButtons(self):
        group = QButtonGroup()
        self.group = group

        button = QRadioButton(None, self)
        button.move(-1000, -1000)  # out of screen
        self.none = button
        group.addButton(button, self.State.NONE)

        label = QLabel(None, self)
        label.setGeometry(0, 422, 200, 28)
        label.setObjectName('button_label')

        button = QRadioButton(None, self)
        button.setObjectName('enter')
        button.move(-5, 253)
        # button.clicked.connect(self.enterClick)
        button.clicked.connect(lambda: self.set_state(self.State.ENTER))
        self.enter = button
        group.addButton(button, self.State.ENTER)

        button.label = label

        #

        label = QLabel(None, self)
        label.setGeometry(200, 422, 200, 28)
        label.setObjectName('button_label')

        button = QPushButton(None, self)
        button.setObjectName('balance')
        button.move(223, 285)
        button.clicked.connect(lambda: self.set_state(self.State.PRE_BALANCE))
        # button.clicked.connect(self.balanceClick)
        self.btn_balance = button

        button.label = label

        #

        label = QLabel(None, self)
        label.setGeometry(400, 422, 200, 28)
        label.setObjectName('button_label')

        button = QPushButton(None, self)
        button.setObjectName('absence')
        button.move(426, 285)
        button.clicked.connect(lambda: self.set_state(self.State.WAIT_FOR_REASON))
        # button.clicked.connect(self.absenceClick)
        self.absence = button

        button.label = label

        #

        label = QLabel(None, self)
        label.setGeometry(600, 422, 200, 28)
        label.setObjectName('button_label')

        button = QRadioButton(None, self)
        button.setObjectName('exit')
        button.move(589, 253)
        # button.clicked.connect(self.exitClick)
        button.clicked.connect(lambda: self.set_state(self.State.EXIT))

        self.exit = button
        group.addButton(button, self.State.EXIT)

        button.label = label

    def alertMsg(self, text):
        self.alert_msg.setText(text)
        self.alert_msg.show()
        self.info_msg.hide()

    def infoMsg(self, text):
        self.info_msg.setText(text)
        self.alert_msg.hide()
        self.info_msg.show()

    def clockMsg(self, text=None, keep_info_on_screen=False):
        if text is not None:
            self.clock_msg.setText(text)
        if not keep_info_on_screen:
            self.alert_msg.hide()
            self.info_msg.hide()

    def passInfo(self, state, name_etc, text=None):
        if state == self.State.ENTER:
            name_etc += '<br>'+_("Вход")
        elif state == self.State.EXIT:
            name_etc += '<br>'+_("Выход")

        name_etc += " " + self.clock_wdg.time.toString('hh:mm')

        if text:
            name_etc += '<br>' + text

        if self.time_violation and self.second_passage:
            alert = _("Повторный проход<br>с нарушением времени")
        elif self.time_violation and self.place_violation:
            alert = _("Регистрация с нарушением<br>зональности и времени")
        elif self.time_violation:
            alert = _("Регистрация с нарушением<br>времени")
        elif self.second_passage:
            alert = _("Повторный проход")
        elif self.place_violation:
            alert = _("Регистрация с нарушением<br>зональности")
        else:
            alert = None

        if alert:
            self.alertMsg(name_etc + "<p><font size=-2><b>"
                          + alert + "</b></font></p>")
        else:
            self.infoMsg(name_etc)

    def setLang(self):
        self.enter.label.setText(_("Вход"))
        self.exit.label.setText(_("Выход"))
        self.btn_balance.label.setText(_("Баланс"))
        self.absence.label.setText(_("Отсутствие"))

    def setReason(self, text):
        self.reason_text = text
        if text is not None:
            self.set_state(self.State.REASON)
        else:
            # impossible?
            self.go_to_default_state()

    def _on_socket_disconnect(self):
        self.clockMsg()  # ??? Надо ли вообще хоть что-тоделать?

    def _on_socket_connect(self):
        if self.state in (self.State.NONE, self.State.ENTER, self.State.EXIT):
            self.protocol.defaultDir(
                self.CONVERT_STATE_TO_DIR[self.state])
        if self.state in (self.State.WAIT_FOR_REASON, self.State.REASON):
            self.protocol.defaultDir(
                self.CONVERT_STATE_TO_DIR[self.reason_state])
        self.go_to_default_state()

    def _on_event_answer(self, msg):
        if 'code' not in msg:
            return

        self.card_waiting_timeout.stop()

        code = msg['code']

        if code == EventCode.INDICATION_EVENT_OFF:
            self.clockMsg()
            return
        if self.state in (self.State.NONE, self.State.ASK_TO_CHOOSE):
            self.state = self.State.ASK_TO_CHOOSE
            self.protocol.unlock('ignore', msg)
            return

        if code not in (EventCode.PASS_PERSONAL, EventCode.IDENTIFY_FP_VERIFY):
            if self.state in (
                    self.State.ENTER, self.State.EXIT, self.State.REASON):
                self.protocol.unlock('pass', msg)
            elif self.state == self.State.PRE_MENU:
                self.protocol.unlock('menu_attempt', msg)
            else:
                self.protocol.unlock('ignore', msg)
            self.reason_text = None

        if code == EventCode.IDLOCK_NOTINLIST:
            self.alertMsg(_("Данные не распознаны<br>попробуйте снова"))
            if self.state != self.State.MENU:
                self.card_waiting_timeout.start()
        elif code == EventCode.IDLOCK_NOVALID:
            self.alertMsg(_("Пропуск не валиден"))
            if self.state != self.State.MENU:
                self.go_to_default_state(keep_info_on_screen=True)
        elif code == EventCode.IDENTIFY_FP_VERIFY:
            self.clockMsg(_("Поднесите палец"))
            # timeout on s30 side -- no need for our timeout (now)
        elif code in (EventCode.IDENTIFY_FP_VERIFY_FAIL,
                      EventCode.IDENTIFY_FAIL):
            text = ''
            if 'identify_violation' in msg:
                violation = msg['identify_violation']
                if violation == 'no_templates':
                    text = _("Нет отпечатков<br>пользователя в БД")
                elif violation == 'unknown_finger':
                    text = _("Неизвестный палец")
                elif violation == 'no_card':
                    text = _("Нет карты")
            self.alertMsg(
                '<font color=white>' + _("Ошибка идентификации") + '<br>'
                + text + '</font>')
            if self.state != self.State.MENU:
                self.go_to_default_state(keep_info_on_screen=True)

        elif code == EventCode.PASS_PERSONAL:
            if self.state == self.State.PRE_MENU:
                master = 'master' in msg and msg['master']
                # self.user_acct = None
                if master:
                    self.state = self.State.MENU
                    # self.protocol.defaultDir(0)
                    # self.menu = MenuWidget(self.protocol, self)
                    self.menu.show()
                    self.protocol.unlock('menu_enter', msg)
                else:
                    self.protocol.unlock('menu_attempt', msg)
                    self.go_to_default_state()
            elif self.state == self.State.PRE_BALANCE:
                self.protocol.unlock('ignore', msg)
                self.state = self.State.NONE
                self.state = self.State.BALANCE
                self.clockMsg(_("Выполняется запрос"))
                self.protocol.getUserInfo(msg['uid'])
                self.protocol.getUserAcct(msg['uid'])
            elif self.state in (
                    self.State.ENTER, self.State.EXIT, self.State.REASON):
                self.time_violation = (
                    'time_violation' in msg and msg['time_violation'])
                self.place_violation = (
                    'place_violation' in msg and msg['place_violation'])
                self.second_passage = (
                    'second_passage' in msg and msg['second_passage'])
                self.protocol.getUserAcct(msg['uid'])
                print('   ### I play! ###')
                if self.time_violation:
                    self.protocol.play('access_prohibited.wav')
                elif self.place_violation:
                    self.protocol.play('access_prohibited.wav')
                elif self.second_passage:
                    self.protocol.play('access_prohibited.wav')
                # elif 'reason' not in msg:
                else:
                    self.protocol.play('access_permit.wav')

                self.protocol.unlock('pass', msg, self.reason_text)
            else:
                self.protocol.unlock('ignore', msg)

    def userAnswer(self, msg):
        user = msg['user']

        if self.state in (self.State.ENTER, self.State.EXIT):
            if 'name' in user:
                self.passInfo(self.state, user['name'])
            self.go_to_default_state(keep_info_on_screen=True)

        elif self.state is self.State.REASON:
            if 'name' in user:
                self.passInfo(
                    self.reason_state, user['name'], self.reason_text)
            self.go_to_default_state(keep_info_on_screen=True)

        elif self.state in (self.State.PRE_BALANCE, self.State.BALANCE):
            self.balance.selected_user = user

    def protocolAnswer(self, msg):
        if 'answer' in msg:
            answer = msg['answer']

            if 'settings' in answer and answer['settings'] == 'ok':
                settings = msg['settings']

                if 'licon_light' in settings:
                    self.brightness.set(settings['licon_light'])

                if 'licon_display_test' in settings \
                        and settings['licon_display_test'] == 'yes':

                    tst_disp = QLabel(self)
                    tst_disp.setGeometry(0, 0, 800, 480)
                    tst_disp.show()
                    tst_disp.raise_()
                    tst_disp_r_timer = QTimer(self)
                    tst_disp_r_timer.timeout.connect(
                        lambda: tst_disp.setStyleSheet(
                            '* { background-color: rgba(0, 255, 0, 255); }'))
                    tst_disp_r_timer.setInterval(1000 * 4 * 1)
                    tst_disp_r_timer.setSingleShot(True)
                    tst_disp_r_timer.start()
                    tst_disp_g_timer = QTimer(self)
                    tst_disp_g_timer.timeout.connect(
                        lambda: tst_disp.setStyleSheet(
                            '* { background-color: rgba(0, 0, 255, 255); }'))
                    tst_disp_g_timer.setInterval(1000 * 4 * 2)
                    tst_disp_g_timer.setSingleShot(True)
                    tst_disp_g_timer.start()
                    tst_disp_b_timer = QTimer(self)
                    tst_disp_b_timer.timeout.connect(tst_disp.hide)
                    tst_disp_b_timer.setInterval(1000 * 4 * 3)
                    tst_disp_b_timer.setSingleShot(True)
                    tst_disp_b_timer.start()
                    tst_disp.setStyleSheet(
                        '* { background-color: rgba(255, 0, 0, 255); }')

            elif 'licon_time' in answer and answer['licon_time'] == 'ok':
                old_default_state = self.state_default
                if 'enter' in msg['licon_time']:
                    self.state_default = self.State.ENTER
                elif 'exit' in msg['licon_time']:
                    self.state_default = self.State.EXIT
                else:
                    self.state_default = self.State.NONE
                if self.state == old_default_state \
                        and self.state != self.state_default:
                    self.go_to_default_state()
            elif 'user_info' in answer and self.state == self.State.BALANCE:
                if answer['user_info'] == 'ok' and 'user_info' in msg:
                    self.balance.show(msg['user_info'])
                else:
                    self.alertMsg(_("Отсутствие связи с сервером"))
                    self.card_waiting_timeout.start()
