import json
import time
import os

from PyQt5.QtCore import QThread, pyqtSignal
from websocket_server import WebsocketServer


class Protocol(QThread):

    events = pyqtSignal(dict)
    users = pyqtSignal(dict)
    answer = pyqtSignal(dict)
    connected = pyqtSignal()
    disconnected = pyqtSignal()

    def __init__(self, port):
        QThread.__init__(self)
        server = WebsocketServer(port)
        server.set_fn_new_client(self._on_connect)
        server.set_fn_client_left(self._on_disconnect)
        server.set_fn_message_received(self.message)
        self.server = server
        self.ready = False
        self.last_msg = None

    def run(self):
        self.server.run_forever()

    # Callbacks #

    def _on_connect(self, client, server):
        print('Сокет (%d) подключен' % client['id'])
        time.sleep(.500)
        self.ready = True
        self._send({'set': 'event_acl',
                    'event_acl': [{'category': 0, 'allow': True}]
                    })
        self._send({'get': 'settings', 'settings': {'prefix': 'time_zone'}})
        self._send({'get': 'settings', 'settings': {'prefix': 'licon'}})
        self.connected.emit()

    def _on_disconnect(self, client, server):
        self.ready = False
        if client is None:
            print('Сокет нештатно отключён, номер сокета неизвестен')
        else:
            print('Сокет (%d) отключен' % client['id'])
        self.disconnected.emit()

    def message(self, client, server, data):
        log = data
        logend = None

        if len(log) > 128:
            log = log[0:128] + '...'
            logend = '...' + data[-128:-1]

        print('WebSock(%d): %s' % (client['id'], log))
        if logend:
            print('WebSock(%d): %s' % (client['id'], logend))
        if not data.strip():
            return
        msg = json.loads(data, strict=False)

        if 'answer' in msg:
            answer = msg['answer']

            if 'settings' in answer:
                if answer['settings'] == 'ok':
                    settings = msg['settings']

                    if 'time_zone' in settings:
                        self.time_zone(msg['settings'])
                    else:
                        self.answer.emit(msg)

            elif 'user' in answer:
                if answer['user'] in ('ok', 'done', 'empty') \
                        and 'set' not in msg['answer']:
                    self.users.emit(msg)
            else:
                if 'set' in answer and answer['set'] == 'foreign':
                    return
                self.answer.emit(msg)

        elif 'event' in msg:
            self.event(msg['event'], client)
        else:
            self.answer.emit(msg)

    # Internal #

    def time_zone(self, msg):
        os.environ['TZ'] = msg['time_zone']

    def event(self, msg, client):
        self.events.emit(msg)

        if 'code' in msg and 'number' in msg:
            self.last_msg = msg

    def _send(self, data):
        if self.ready:
            msg = json.dumps(data, ensure_ascii=False)
            print('WebSock(*): %s' % (msg))
            self.server.send_message_to_all(msg + '\n\n')

    # External #

    def unlock(self, passage=False, msg=None, reason=None):
        cmd = {'action': 'unlock', 'place': 'all', 'gateway_dir': 0}

        if msg is not None and 'number' in msg:
            cmd['exdev'] = msg['number']
        else:
            if self.last_msg:
                cmd['exdev'] = self.last_msg['number']
            else:
                return

        if msg is not None and 'dir' in msg:
            cmd['number'] = msg['dir']
        else:
            if self.last_msg:
                cmd['number'] = self.last_msg['dir']
            else:
                return

        if msg is not None and 'uid' in msg:
            cmd['user_acct'] = msg['uid']
        elif self.last_msg is not None and 'uid' in self.last_msg:
            if self.last_msg:
                cmd['user_acct'] = self.last_msg['uid']
            else:
                return

        if passage is not None:
            cmd['user_action'] = passage

        if reason is not None:
            cmd['reason'] = reason

        self._send({'control': 'direction', 'direction': cmd})

    def defaultDir(self, num):
        self._send({
            'control': 'direction',
            'direction': {'exdev': 0, 'number': num, 'action': 'set_default'}})

    def reason(self, text, acct):
        self._send({
            'control': 'reason',
            'reason': {'text': text, 'user_acct': acct}})

    def manual(self, id):
        self._send({'control': 'manual', 'manual': {'id': id}})

    def getUserName(self, name):
        self._send({
            'get': 'user',
            'user': {'name': name + '%', 'limit': 10, 'done': True}})

    def getUserAcct(self, acct):
        self._send({'get': 'user', 'user': {'acct': acct}})

    def getUserInfo(self, acct):
        self._send({'get': 'user_info', 'user_info': {'user_acct': acct}})

    def newCard(self, acct, value):
        self._send({'set': 'uid', 'uid': {'user_acct': acct, 'value': value}})

    def scanOnFinger(self, acct, num, quality):
        self._send({
            'control': 'finger',
            'finger': {
                'scan': True, 'image': True,
                'user_acct': acct, 'finger_num': num, 'quality': quality}})

    def scanOffFinger(self):
        self._send({'control': 'finger', 'finger': {'scan': False}})

    def clearFinger(self, acct):
        self._send({'set': 'finger', 'finger': {'user_acct': acct}})

    def getFinger(self, acct):
        self._send({
            'get': 'finger',
            'finger': {'user_acct': acct, 'done': True}})

    def deleteCard(self, value):
        self._send({'set': 'uid', 'uid': {'value': value}})

    def getCard(self, acct):
        self._send({'get': 'uid', 'uid': {'user_acct': acct, 'done': True}})

    def setSysTime(self, time):
        self._send({'set': 'systime', 'systime': {'time': int(time)}})

    def setSettings(self, values):
        self._send({'set': 'settings', 'settings': values})

    def getSettings(self, name):
        self._send({'get': 'settings', 'settings': {'prefix': name}})

    def play(self, name):
        self._send({'control': 'sound', 'sound': {'name': 'sound/' + name}})

    def getSysinfo(self):
        self._send({'get': 'sysinfo'})

    def getDefault(self):
        self._send({'get': 'licon_time'})
